==> chasen-sys2.0b6-ipadic1.0.tar.gz: chasen-2.0b6/lib/connect.c <==
/*
==============================================================================
	connect.c
		1990/12/17/Mon	Yutaka MYOKI(Nagao Lab., KUEE)
==============================================================================
*/

#include "chadic.h"

#define CHA_LINEMAX 8192

static int tbl_num;          /* ϢɽΥ */
static int tbl_num_goi;	     /* Ϣɽθäο */
static int i_num;            /* Ϣܹι   */
static int j_num;            /* Ϣܹ   */
static rensetu_pair_t *rensetu_tbl;
static connect_rule_t *connect_mtr;

/*
------------------------------------------------------------------------------
        rensetu table
------------------------------------------------------------------------------
*/
static int cmp_pair(pair1, pair2)
    rensetu_pair_t *pair1, *pair2;
{
    int ret;
    
    /* Ф */
    if (pair1->goi == NULL && pair2->goi != NULL)
      return 1;
    if (pair1->goi != NULL && pair2->goi == NULL)
      return -1;
    
    /* ʻʬ */
    if (ret = pair1->hinsi - pair2->hinsi)
      return ret;

#ifndef VGRAM
    /* ʻʬ */
    if (ret = pair1->bunrui - pair2->bunrui)
      return ret;
#endif
    
    /* ѷ*/
    if (ret = pair1->type - pair2->type)
      return ret;
    
    /* Ф */
    if (pair1->goi != NULL && pair2->goi != NULL)
      if (ret = strcmp(pair1->goi, pair2->goi))
	return ret;
    
    /* ѷ */
    if (ret = pair1->form - pair2->form)
      return ret;
    
    return pair1->index - pair2->index;
}

static char *cha_numtok(s, valp)
    char *s;
    int *valp;
{
    int minus;

    while (*s == ' ')
      s++;
    if (*s != '-')
      minus = 0;
    else {
	minus = 1;
	s++;
    }
    if (*s < '0' || *s > '9')
      cha_exit_file(1, "illegal format");
    for (*valp = 0; *s >= '0' && *s <= '9'; *valp = *valp * 10 + *s++ - '0');
    while (*s == ' ')
      s++;

    if (minus)
      *valp = -*valp;

    return s;
}

void read_table(fp_out, dir)
    FILE *fp_out;
    int dir;
{
    FILE *fp;
    char *filepath;
    char buf[CHA_LINEMAX], *s;
    int i, val, cell_num;

#ifdef VGRAM
    fp = cha_fopen_grammar(TABLEFILE, "r", 1, dir, &filepath);
#else
    fp = cha_fopen_grammar2(TABLEFILE, JM_TABLEFILE, "r", 1, dir, &filepath);
#endif

    if (fp_out != NULL)
      fprintf(fp_out, "parsing %s\n", filepath);

    LineNoForError = ++LineNo;
    fscanf(fp, "%d\n", &cell_num);
    rensetu_tbl =
      (rensetu_pair_t *)cha_malloc(sizeof(rensetu_pair_t) * cell_num);

    tbl_num = 0;
    for (i = 0; i < cell_num; i++) {
	LineNoForError = ++LineNo;
	if (fgets(buf, sizeof(buf), fp) == NULL)
	  cha_exit_file(1, "illegal format");
	LineNoForError = ++LineNo;
	if (fgets(s = buf, sizeof(buf), fp) == NULL)
	  cha_exit_file(1, "illegal format");
	s = cha_numtok(s, &val);
	rensetu_tbl[i].i_pos = val;
	s = cha_numtok(s, &val);
	rensetu_tbl[i].j_pos = val;
	if (!tbl_num && val < 0)
	  tbl_num = i;
	buf[strlen(buf) - 1] = '\0';
	if (*s >= '0' && *s <= '9') {
	    s = cha_numtok(s, &val);
	    rensetu_tbl[i].index = i;
	    rensetu_tbl[i].hinsi = val;
#ifndef VGRAM
	    s = cha_numtok(s, &val);
	    rensetu_tbl[i].bunrui = val;
#endif
	    s = cha_numtok(s, &val);
	    rensetu_tbl[i].type = val;
	    s = cha_numtok(s, &val);
	    rensetu_tbl[i].form = val;
	    if (*s == '*') {
		rensetu_tbl[i].goi = NULL;
	    } else {
		rensetu_tbl[i].goi = cha_strdup(s);
		tbl_num_goi++;
	    }
	} else { /* for EDRdic '94.Mar */
	    rensetu_tbl[i].hinsi = UCHAR_MAX;
	    rensetu_tbl[i].goi = cha_strdup(s);
	}
    }

    if (!tbl_num)
      tbl_num = cell_num;

    fclose(fp);
}

/* for EDRdic '94.Mar */
void check_edrtable(mrph_p, x)
    mrph_t *mrph_p;
    cell_t *x;
{
    int i;

    for (i = 0; i < tbl_num; i++) {
	if ( rensetu_tbl[i].hinsi == UCHAR_MAX &&
	    strcmp(s_atom(x), rensetu_tbl[i].goi) == 0) {
	    mrph_p->con_tbl = i;
	    return;
	}
    }
    cha_exit_file(1, "no morpheme in EDR table");
}

/* for EDRdic '94.Mar */
void check_edrtable_str(mrph_p, str)
    mrph_t *mrph_p;
    char *str;
{
    int i;
  
    for (i = 0; i < tbl_num; i++) {
	if ( rensetu_tbl[i].hinsi == UCHAR_MAX &&
	    strcmp(str, rensetu_tbl[i].goi) == 0) {
	    mrph_p->con_tbl = i;
	    return;
	}
    }
    cha_exit_file(1, "no morpheme in EDR table");
}

static int find_table(mrph, pair)
    mrph_t *mrph;
    rensetu_pair_t *pair;
{
    int ret;

    /* ʻʬ */
    if (ret = mrph->hinsi - pair->hinsi)
      return ret;
#ifndef VGRAM
    /* ʻʬ */
    if (ret = mrph->bunrui - pair->bunrui)
      return ret;
#endif
    /* ѷ */
    if (ret = mrph->ktype - pair->type)
      return ret;

    /* Ф */
    if (pair->goi && (ret = strcmp(mrph->midasi,pair->goi)))
      return ret;

    /* ѸʤСѷ1֤ȥޥå*/
    if (mrph->ktype)
      return 1 - pair->form;
    return 0;
}

int check_table(mrph_p) /* 970301 tatuo: void -> int for  */
    mrph_t *mrph_p;
{
    int i, hinsi;
    rensetu_pair_t *ret;

    if (rensetu_tbl[0].hinsi == 0)
      qsort(rensetu_tbl, tbl_num, sizeof(rensetu_pair_t), (int (*)())cmp_pair);

    ret = (rensetu_pair_t *)
      bsearch(mrph_p, rensetu_tbl, tbl_num_goi,
	      sizeof(rensetu_pair_t), (int (*)())find_table);
    if (ret){
	mrph_p->con_tbl = ret->index;
	return 1; /* 970301 tatuo: ʤ 1 ֤ */
    }

    ret = (rensetu_pair_t *)
      bsearch(mrph_p, rensetu_tbl + tbl_num_goi, tbl_num - tbl_num_goi,
	      sizeof(rensetu_pair_t), (int (*)())find_table);
    if (ret){
	mrph_p->con_tbl = ret->index;
	return 1; /* 970301 tatuo: ʤ 1 ֤ */
    }

    cha_exit_file(-1, "no morpheme in table"); /* 970301 tatuo: 1 -> -1 */
    return 0; /* 970301 tatuo: ꤢ 0 ֤ */
}

#ifdef VGRAM
int check_table_for_undef(hinsi)
    int hinsi;
{
     int i;

     for (i = 0; i < tbl_num; i++)
       if (rensetu_tbl[i].hinsi == hinsi)
	 if (!rensetu_tbl[i].goi)
	   return i;

     return -1;
}
#else
int check_table_for_undef(hinsi, bunrui)
    int hinsi, bunrui;
{
     int i;

     for (i = 0; i < tbl_num; i++)
       if (rensetu_tbl[i].hinsi == hinsi && rensetu_tbl[i].bunrui == bunrui)
	 if (!rensetu_tbl[i].goi) /* bug fixed by A.Kitauchi */
	   return i;

     return -1;
}
#endif

/*
------------------------------------------------------------------------------
        rensetu matrix
------------------------------------------------------------------------------
*/
void read_matrix(fp_out)
    FILE *fp_out;
{
    FILE *fp;
    char *filepath;
    int i, j, cost, next;
    char buf[CHA_LINEMAX], *s;

#ifdef VGRAM
    fp = cha_fopen_grammar(MATRIXFILE, "r", 1, 1, &filepath);
#else
    fp = cha_fopen_grammar2(MATRIXFILE, JM_MATRIXFILE, "r", 1, 1, &filepath);
#endif

    if (fp_out != NULL)
      fprintf(fp_out, "parsing %s", filepath);

    LineNoForError = ++LineNo;
    fscanf(fp, "%d %d\n", &i_num, &j_num);
    connect_mtr = (connect_rule_t *)
      cha_malloc(sizeof(connect_rule_t) * i_num * j_num);

    next = 0;
    for (i = 0; i < i_num; i++) {
	LineNoForError = ++LineNo;
	if (fgets(s = buf, sizeof(buf), fp) == NULL)
	  cha_exit_file(1, "illegal format");
	for (j = 0; j < j_num; j++) {
#ifdef VGRAM
	    s = cha_numtok(s, &next);
	    if (*s++ != ',')
	      cha_exit_file(1, "illegal format");
#endif
	    s = cha_numtok(s, &cost);
	    connect_mtr[i * j_num + j].next = next;
	    connect_mtr[i * j_num + j].cost = cost;
	}
    }
    fclose(fp);
}

#ifdef KOCHA
int check_automaton(state, state_prev, midasi, is_undef, con, undef_con_cost, costp)
    int state, state_prev;
    char *midasi;
    int is_undef, con, undef_con_cost, *costp;
{
    int cost, cost_prev;

    cost = connect_mtr[state * j_num + rensetu_tbl[con].j_pos].cost;
    /* 2: ľñȤ³ */
    if (cost == 0)
      cost = undef_con_cost;
    else if (!is_undef &&
	     (unsigned char)midasi[0] == 0xA1 &&
	     (unsigned char)midasi[1] == 0xA1) {
	/* 2: 2ñȤ³ */
	cost_prev =
	  (int)connect_mtr[state_prev * j_num + rensetu_tbl[con].j_pos].cost / 100;

#ifdef KOCHA_DEBUG
	printf("[c1:%d,c2:%d,cost:%d]\n",state_prev,rensetu_tbl[con].j_pos,cost_prev);
#endif
	if (cost_prev == 0)
	  cost = undef_con_cost;
	else
	  cost += cost_prev;
    }

    *costp = cost;

    return rensetu_tbl[con].i_pos;
}
#else /* !KOCHA */
int check_automaton(state, con, undef_con_cost, costp)
    int state, con, undef_con_cost, *costp;
{
    connect_rule_t *cr;
#if 0
    printf("[%d:%d:%d]\n",state,con,rensetu_tbl[con].j_pos);fflush(stdout);
#endif
    cr = &connect_mtr[state * j_num + rensetu_tbl[con].j_pos];
    *costp = cr->cost;
    if (*costp == 0)
      *costp = undef_con_cost;

#ifdef DEBUG
    printf("[state:%d,con:%d,newcon:%d] ",state,con,cr->next+con);
#endif

    return rensetu_tbl[cr->next + con].i_pos;
}
#endif /* !KOCHA */

==> chasen-sys2.01-ipadic2.0.tar.gz: chasen-2.01/lib/connect.c <==
/*
==============================================================================
	connect.c
		1990/12/17/Mon	Yutaka MYOKI(Nagao Lab., KUEE)
==============================================================================
*/

#include "chadic.h"

#define CHA_LINEMAX 8192

static int tbl_num;          /* ϢɽΥ */
static int tbl_num_goi;	     /* Ϣɽθäο */
static int i_num;            /* Ϣܹι   */
static int j_num;            /* Ϣܹ   */
static rensetu_pair_t *rensetu_tbl;
static connect_rule_t *connect_mtr;

/*
------------------------------------------------------------------------------
        rensetu table
------------------------------------------------------------------------------
*/
static int cmp_pair(pair1, pair2)
    rensetu_pair_t *pair1, *pair2;
{
    int ret;
    
    /* Ф */
    if (pair1->goi == NULL && pair2->goi != NULL)
      return 1;
    if (pair1->goi != NULL && pair2->goi == NULL)
      return -1;
    
    /* ʻʬ */
    if (ret = pair1->hinsi - pair2->hinsi)
      return ret;

#ifndef VGRAM
    /* ʻʬ */
    if (ret = pair1->bunrui - pair2->bunrui)
      return ret;
#endif
    
    /* ѷ*/
    if (ret = pair1->type - pair2->type)
      return ret;
    
    /* Ф */
    if (pair1->goi != NULL && pair2->goi != NULL)
      if (ret = strcmp(pair1->goi, pair2->goi))
	return ret;
    
    /* ѷ */
    if (ret = pair1->form - pair2->form)
      return ret;
    
    return pair1->index - pair2->index;
}

static char *cha_numtok(s, valp)
    char *s;
    int *valp;
{
    int minus;

    while (*s == ' ')
      s++;
    if (*s != '-')
      minus = 0;
    else {
	minus = 1;
	s++;
    }
    if (*s < '0' || *s > '9')
      cha_exit_file(1, "illegal format");
    for (*valp = 0; *s >= '0' && *s <= '9'; *valp = *valp * 10 + *s++ - '0');
    while (*s == ' ')
      s++;

    if (minus)
      *valp = -*valp;

    return s;
}

void read_table(fp_out, dir)
    FILE *fp_out;
    int dir;
{
    FILE *fp;
    char *filepath;
    char buf[CHA_LINEMAX], *s;
    int i, val, cell_num;

#ifdef VGRAM
    fp = cha_fopen_grammar(TABLE_FILE, "r", 1, dir, &filepath);
#else
    fp = cha_fopen_grammar2(TABLE_FILE, JM_TABLE_FILE, "r", 1, dir, &filepath);
#endif

    if (fp_out != NULL)
      fprintf(fp_out, "parsing %s\n", filepath);

    LineNoForError = ++LineNo;
    fscanf(fp, "%d\n", &cell_num);
    rensetu_tbl =
      (rensetu_pair_t *)cha_malloc(sizeof(rensetu_pair_t) * cell_num);

    tbl_num = 0;
    for (i = 0; i < cell_num; i++) {
	LineNoForError = ++LineNo;
	if (fgets(buf, sizeof(buf), fp) == NULL)
	  cha_exit_file(1, "illegal format");
	LineNoForError = ++LineNo;
	if (fgets(s = buf, sizeof(buf), fp) == NULL)
	  cha_exit_file(1, "illegal format");
	s = cha_numtok(s, &val);
	rensetu_tbl[i].i_pos = val;
	s = cha_numtok(s, &val);
	rensetu_tbl[i].j_pos = val;
	if (!tbl_num && val < 0)
	  tbl_num = i;
	buf[strlen(buf) - 1] = '\0';
	if (*s >= '0' && *s <= '9') {
	    s = cha_numtok(s, &val);
	    rensetu_tbl[i].index = i;
	    rensetu_tbl[i].hinsi = val;
#ifndef VGRAM
	    s = cha_numtok(s, &val);
	    rensetu_tbl[i].bunrui = val;
#endif
	    s = cha_numtok(s, &val);
	    rensetu_tbl[i].type = val;
	    s = cha_numtok(s, &val);
	    rensetu_tbl[i].form = val;
	    if (*s == '*') {
		rensetu_tbl[i].goi = NULL;
	    } else {
		rensetu_tbl[i].goi = cha_strdup(s);
		tbl_num_goi++;
	    }
	} else { /* for EDRdic '94.Mar */
	    rensetu_tbl[i].hinsi = UCHAR_MAX;
	    rensetu_tbl[i].goi = cha_strdup(s);
	}
    }

    if (!tbl_num)
      tbl_num = cell_num;

    fclose(fp);
}

/* for EDRdic '94.Mar */
void check_edrtable(mrph_p, x)
    mrph_t *mrph_p;
    cell_t *x;
{
    int i;

    for (i = 0; i < tbl_num; i++) {
	if ( rensetu_tbl[i].hinsi == UCHAR_MAX &&
	    strcmp(s_atom(x), rensetu_tbl[i].goi) == 0) {
	    mrph_p->con_tbl = i;
	    return;
	}
    }
    cha_exit_file(1, "no morpheme in EDR table");
}

/* for EDRdic '94.Mar */
void check_edrtable_str(mrph_p, str)
    mrph_t *mrph_p;
    char *str;
{
    int i;
  
    for (i = 0; i < tbl_num; i++) {
	if ( rensetu_tbl[i].hinsi == UCHAR_MAX &&
	    strcmp(str, rensetu_tbl[i].goi) == 0) {
	    mrph_p->con_tbl = i;
	    return;
	}
    }
    cha_exit_file(1, "no morpheme in EDR table");
}

static int find_table(mrph, pair)
    mrph_t *mrph;
    rensetu_pair_t *pair;
{
    int ret;

    /* ʻʬ */
    if (ret = mrph->hinsi - pair->hinsi)
      return ret;
#ifndef VGRAM
    /* ʻʬ */
    if (ret = mrph->bunrui - pair->bunrui)
      return ret;
#endif
    /* ѷ */
    if (ret = mrph->ktype - pair->type)
      return ret;

    /* Ф */
    if (pair->goi && (ret = strcmp(mrph->midasi,pair->goi)))
      return ret;

    /* ѸʤСѷ1֤ȥޥå*/
    if (mrph->ktype)
      return 1 - pair->form;
    return 0;
}

int check_table(mrph_p) /* 970301 tatuo: void -> int for  */
    mrph_t *mrph_p;
{
    int i, hinsi;
    rensetu_pair_t *ret;

    if (rensetu_tbl[0].hinsi == 0)
      qsort(rensetu_tbl, tbl_num, sizeof(rensetu_pair_t), (int (*)())cmp_pair);

    ret = (rensetu_pair_t *)
      bsearch(mrph_p, rensetu_tbl, tbl_num_goi,
	      sizeof(rensetu_pair_t), (int (*)())find_table);
    if (ret){
	mrph_p->con_tbl = ret->index;
	return 1; /* 970301 tatuo: ʤ 1 ֤ */
    }

    ret = (rensetu_pair_t *)
      bsearch(mrph_p, rensetu_tbl + tbl_num_goi, tbl_num - tbl_num_goi,
	      sizeof(rensetu_pair_t), (int (*)())find_table);
    if (ret){
	mrph_p->con_tbl = ret->index;
	return 1; /* 970301 tatuo: ʤ 1 ֤ */
    }

    cha_exit_file(-1, "no morpheme in connection table\n"); /* 970301 tatuo: 1 -> -1 */
    return 0; /* 970301 tatuo: ꤢ 0 ֤ */
}

#ifdef VGRAM
int check_table_for_undef(hinsi)
    int hinsi;
{
     int i;

     for (i = 0; i < tbl_num; i++)
       if (rensetu_tbl[i].hinsi == hinsi)
	 if (!rensetu_tbl[i].goi)
	   return i;

     return -1;
}
#else
int check_table_for_undef(hinsi, bunrui)
    int hinsi, bunrui;
{
     int i;

     for (i = 0; i < tbl_num; i++)
       if (rensetu_tbl[i].hinsi == hinsi && rensetu_tbl[i].bunrui == bunrui)
	 if (!rensetu_tbl[i].goi) /* bug fixed by A.Kitauchi */
	   return i;

     return -1;
}
#endif

/*
------------------------------------------------------------------------------
        rensetu matrix
------------------------------------------------------------------------------
*/
void read_matrix(fp_out)
    FILE *fp_out;
{
    FILE *fp;
    char *filepath;
    int i, j, cost, next;
    char buf[CHA_LINEMAX], *s;

#ifdef VGRAM
    fp = cha_fopen_grammar(MATRIX_FILE, "r", 1, 1, &filepath);
#else
    fp = cha_fopen_grammar2(MATRIX_FILE, JM_MATRIX_FILE, "r", 1, 1, &filepath);
#endif

    if (fp_out != NULL)
      fprintf(fp_out, "parsing %s", filepath);

    LineNoForError = ++LineNo;
    fscanf(fp, "%d %d\n", &i_num, &j_num);
    connect_mtr = (connect_rule_t *)
      cha_malloc(sizeof(connect_rule_t) * i_num * j_num);

    next = 0;
    for (i = 0; i < i_num; i++) {
	LineNoForError = ++LineNo;
	if (fgets(s = buf, sizeof(buf), fp) == NULL)
	  cha_exit_file(1, "illegal format");
	for (j = 0; j < j_num; j++) {
#ifdef VGRAM
	    if (*s == 'o') {
		int nzero;
		s = cha_numtok(s + 1, &nzero);
		while (--nzero > 0) {
		    connect_mtr[i * j_num + j].next = 0;
		    connect_mtr[i * j_num + j].cost = 0;
		    j++;
		}
		next = cost = 0;
	    } else {
		s = cha_numtok(s, &next);
		if (*s++ != ',')
		  cha_exit_file(1, "illegal format");
		s = cha_numtok(s, &cost);
	    }
#else
	    s = cha_numtok(s, &cost);
#endif
	    connect_mtr[i * j_num + j].next = next;
	    connect_mtr[i * j_num + j].cost = cost;
	}
    }
    fclose(fp);
}

#ifdef KOCHA
int check_automaton(state, state_prev, midasi, is_undef, con, undef_con_cost, costp)
    int state, state_prev;
    char *midasi;
    int is_undef, con, undef_con_cost, *costp;
{
    int cost, cost_prev;

    cost = connect_mtr[state * j_num + rensetu_tbl[con].j_pos].cost;
    /* 2: ľñȤ³ */
    if (cost == 0)
      cost = undef_con_cost;
    else if (!is_undef &&
	     (unsigned char)midasi[0] == 0xA1 &&
	     (unsigned char)midasi[1] == 0xA1) {
	/* 2: 2ñȤ³ */
	cost_prev =
	  (int)connect_mtr[state_prev * j_num + rensetu_tbl[con].j_pos].cost / 100;

#ifdef KOCHA_DEBUG
	printf("[c1:%d,c2:%d,cost:%d]\n",state_prev,rensetu_tbl[con].j_pos,cost_prev);
#endif
	if (cost_prev == 0)
	  cost = undef_con_cost;
	else
	  cost += cost_prev;
    }

    *costp = cost;

    return rensetu_tbl[con].i_pos;
}
#else /* !KOCHA */
int check_automaton(state, con, undef_con_cost, costp)
    int state, con, undef_con_cost, *costp;
{
    connect_rule_t *cr;
#if 0
    printf("[%d:%d:%d]\n",state,con,rensetu_tbl[con].j_pos);fflush(stdout);
#endif
    cr = &connect_mtr[state * j_num + rensetu_tbl[con].j_pos];
    *costp = cr->cost;
    if (*costp == 0)
      *costp = undef_con_cost;
#ifdef VGRAM
    else
      (*costp)--;
#endif

#ifdef DEBUG
    printf("[state:%d,con:%d,newcon:%d] ",state,con,cr->next+con);
#endif

    return rensetu_tbl[cr->next + con].i_pos;
}
#endif /* !KOCHA */
